;+
; NAME:
;    rastr
; PURPOSE: (one line)
;    Convert RA in radians to hours, minutes, and seconds (ASCII string).
; DESCRIPTION:
;
; CATEGORY:
;    Astronomy
; CALLING SEQUENCE:
;    rastr, radians, places, str, carry
; INPUTS:
;    ra     : Right Ascension, in radians, to be converted to a string.
;             May a vector, in which case the outputs will be vectors.
;    places : Resolution of output string (integer).
;             = -8     Decimal hours                     HH.hhhhh
;             = -7     nearest hour.                     HH
;             = -6     nearest 30 minutes.               HH:(00,30)
;             = -5     nearest 15 minutes.               HH:(00,15,30,45)
;             = -4     nearest 10 minutes.               HH:M0
;             = -3     nearest 5 minutes.                HH:M(0,5)
;             = -2     nearest minute.                   HH:MM
;             = -1     nearest ten seconds.              HH:MM:S0
;             =  0     nearest second.                   HH:MM:SS
;             =  1     nearest tenth of a second.        HH:MM:SS.s
;             =  2     nearest hundredth of a second.    HH:MM:SS.ss
;             =  3     nearest thousandth of a second.   HH:MM:SS.sss
;             =  4     nearest thousandth of a second.   HH:MM:SS.ssss
; OPTIONAL INPUT PARAMETERS:
;
; KEYWORD PARAMETERS:
;
; OUTPUTS:
;    str   - Output string for the converted right ascension.
;    carry - Flag that indicates ra rolled over after rounding.
; COMMON BLOCKS:
;
; SIDE EFFECTS:
;
; RESTRICTIONS:
;
; PROCEDURE:
;    Right ascension is reduced to the range [0,24) then converted to a string.
;    There are no imbedded blanks.
;    Calls external routine radtohms.
; MODIFICATION HISTORY:
;  Copyright (C) 1987, by Marc W. Buie
;  Version dated 87/4/21
;    Ported by Doug loucks, Lowell Observatory, July, 1993, from the
; C-Language version written by Marc Buie.
; 97/07/14, MWB, added places=4
;  99/1/12, MWB, added places = -8
;-
; ------------------------------------------------------------------------------
; Local procedure dadjust
; ------------------------------------------------------------------------------
PRO dadjust, hh, mm, carry
i = WHERE( mm EQ 60, count )
IF count GT 0 THEN BEGIN
   hh[i] = hh[i] + 1
   mm[i] =  0
   j = WHERE( hh[i] EQ 24, count )
   IF count GT 0 THEN BEGIN
      hh[i[j]] = 0
      carry[i[j]] = 1
   ENDIF
ENDIF
END

; ------------------------------------------------------------------------------
; Local procedure madjust
; ------------------------------------------------------------------------------
PRO madjust, hh, mm, ss, carry
i = WHERE( ss GE 60.0, count )
IF count GT 0 THEN BEGIN
   mm[i] = mm[i] + 1
   ss[i] =  0
   j = WHERE( mm[i] EQ 60, count )
   IF count GT 0 THEN BEGIN
      mm[i[j]] = 0
      hh[i[j]] = hh[i[j]] + 1
      k = WHERE( hh[i[j]] EQ 24, count )
      IF count GT 0 THEN BEGIN
         hh[i[j[k]]] = 0
         carry[i[j[k]]] = 1
      ENDIF
   ENDIF
ENDIF
END

; ------------------------------------------------------------------------------
; Main procedure rastr
; ------------------------------------------------------------------------------
PRO rastr, ra, places, str, carry
IF N_PARAMS() LT 3 THEN BEGIN
   ; Display the calling sequence.
   PRINT, 'rastr, radians, places, str, carry'
   RETURN
ENDIF

; Allow input radians to be integer, long, floating, or double, and scalar
; or vector.
IF badpar( ra, [2,3,4,5], [0,1], CALLER='RASTR: (ra) ' ) THEN RETURN

; Allow places to be integer or long scalar.
IF badpar( places, [2,3], 0, CALLER='RASTR: (places) ' ) THEN RETURN

true  = 1
false = 0

; First, convert to numerical values of hours, minutes, and seconds.
radtohms, ra, hh, mm, ss

; Initialize the carry parameter.  This statement, although not the most
; efficient, automatically sets 'carry' to have the same rank as 'ra'.
carry = FIX( ra ) * 0

; This is the big decision table that generates the requested format.
CASE places OF
   -8 : BEGIN   ; HH.hhhhh, Decimal hours
      hh = double(hh) + double(mm)/60.0d0 + ss/3600.0d0
      hh = double(long(hh * 100000.0d0 + 0.5d0))/100000.0d0
      fhh = hh - fix( hh )
      hh  = fix(hh)
      i = where( hh eq 24, count)
      if count gt 0 then begin
         hh[i] = hh[i]-24
         carry[i] = true
      endif
      str = string( hh, format='(i2.2)' ) + $
            STRMID( string( fhh, format='(F7.5)' ), 1, 6 )
   END

   -7 : begin   ; HH, Nearest hour.
      i = WHERE( float(mm) + ss/60.0 GE 30.0, count )
      IF count GT 0 then hh[i]=hh[i]+1
      j = WHERE( hh[i] EQ 24, count )
      IF count GT 0 then begin
         hh[i[j]] = 0
         carry[i[j]] = true
      endif
      str = string( hh, FORMAT='(I2.2)' )
   END

   -6 : begin  ; HH:(00,30), Nearest 30 minutes.
      mm = nint( ( mm + ss / 60.0 ) / 30.0 ) * 30.0
      dadjust, hh, mm, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' )
   END

   -5 : begin  ; HH:(00,15,30,45), Nearest 15 minutes.
      mm = nint( ( mm + ss / 60.0 ) / 15.0 ) * 15.0
      dadjust, hh, mm, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' )
   END

   -4 : begin  ; HH:M0, Nearest 10 minutes.
      mm = nint( ( mm + ss / 60.0 ) / 10.0 ) * 10.0
      dadjust, hh, mm, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' )
   END

   -3 : begin   ; HH:M(0,5), Nearest 5 minutes.
      mm = nint( ( mm + ss / 60.0 ) / 5.0 ) * 5.0
      dadjust, hh, mm, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' )
   END

   -2 : begin    ; HH:MM, Nearest minute.
      mm = nint( mm + ss / 60.0 )
      dadjust, hh, mm, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' )
   END

   -1 : begin  ; HH:MM:S0, Nearest 10 seconds.
      ss = nint( ss / 10.0 ) * 10.0
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' )
   END

   0 : begin    ; HH:MM:SS, Nearest second.
      ss = nint( ss )
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' )
   END

   1 : begin  ; HH:MM:SS.s, Nearest tenth of a second.
      ss = nint( ss / 0.1 ) * 0.1
      ; Save the fractional part of the seconds.
      fss = ss - FIX( ss )
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' ) + $
            STRMID( string( fss, FORMAT='(F3.1)' ), 1, 2 )
   END

   2 : begin  ; HH:MM:SS.ss, Nearest hundredth of a second.
      ss = nint( ss / 0.01 ) * 0.01
      ; Save the fractional part of the seconds.
      fss = ss - FIX( ss )
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' ) + $
            STRMID( string( fss, FORMAT='(F4.2)' ), 1, 3 )
   END

   3 : begin  ; HH:MM:SS.sss, Nearest thousandth of a second.
      ss = nint( ss / 0.001, /LONG ) * 0.001
      ; Save the fractional part of the seconds.
      fss = ss - FIX( ss )
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' ) + $
            STRMID( string( fss, FORMAT='(F5.3)' ), 1, 4 )
   END

   4 : begin  ; HH:MM:SS.ssss, Nearest ten thousandth of a second.
      ss = nint( ss / 0.0001, /LONG ) * 0.0001
      ; Save the fractional part of the seconds.
      fss = ss - FIX( ss )
      madjust, hh, mm, ss, carry
      str = string( hh, FORMAT='(I2.2)' ) + ':' + $
            string( mm, FORMAT='(I2.2)' ) + ':' + $
            string( ss, FORMAT='(I2.2)' ) + $
            STRMID( string( fss, FORMAT='(F6.4)' ), 1, 5 )
   END

   ELSE : begin
      MESSAGE, 'PLACES parameter must be in the range -8 to +4.', /INFO
      RETURN
   END
ENDCASE
END
