pro rdphalt,logname,filename,obj,fil,jd,exptime,gain,rad,sky1,sky2,serial, $
           xpos,ypos,fwhm,maxcnt,mag,err,bad,SKIPLINE=skipline,NUMLINE=numline
;+
; NAME:
;	rdphalt
; PURPOSE: (one line)
;	Reads photometry from an alternate format basphot log file.
; DESCRIPTION:
;	All photometry is read from the file and returned to the caller.
; CATEGORY:
;  File I/O
; CALLING SEQUENCE:
;	pro rdphalt,logname,filename,obj,fil,jd,exptime,gain,rad,sky1,sky2,serial, $
;	           xpos,ypos,fwhm,maxcnt,mag,err
; INPUTS:
;	logname  - string containing the photometry log file to read from.
; OPTIONAL INPUT PARAMETERS:
; KEYWORD PARAMETERS:
; OUTPUTS:
;	filename - Original image file name for each point.
;	obj      - String array containing all object names.
;	fil      - String array with filter codes.
;	jd       - Julian date of mid-time of observation.
;	exptime  - Exposure time in seconds.
;	gain     - Gain of system in photons per count.
;	rad      - Radius of object aperture in pixels.
;	sky1     - Inner radius of sky annulus in pixels.
;	sky2     - Outer radius of sky annulus in pixels.
;	serial   - Serial number for observation.
;	xpos     - vector containing the x position for each point.
;	ypos     - vector containing the y position for each point.
;	fwhm     - Full-width at half maximum in arcseconds.
;	maxcnt   - Original DN of maximum in image.
;	mag      - Instrumental magnitude.
;	err      - Uncertainties of the magnitudes.
;  bad      - flag that marks data as bad.  1 is bad, 0 or missing is good.
; COMMON BLOCKS:
; SIDE EFFECTS:
; RESTRICTIONS:
; PROCEDURE:
; MODIFICATION HISTORY:
;	Written 1/19/93 - Marc W. Buie, Lowell Observatory.
;	2/4/93, MWB, Rewrote file read to eliminate array concatenation.
;  96/01/11, MWB, added NUMLINE, SKIPLINE keywords
;  96/10/31, MWB, added bad flag argument and field.
;-

on_ioerror,badio

if not exists(logname) then begin
   print,'RDPHOTALT: File ',logname,' not found.  Aborting.'
   return
endif

if badpar(skipline,[0,2,3],0,caller='RDPHALT: (skipline) ',default=0) then return
if badpar(numline,[0,2,3],0,caller='RDPHALT: (numline) ',default=-1) then return

openr,unit,logname,/get_lun

line=''

;Read through and count the number of lines.
nobs=0
while(not eof(unit)) do begin
   readf,unit,line,format='(a1)'
   nobs=nobs+1
endwhile
nlines=nobs

if skipline gt nobs then begin
   print,'RDPHALT: skipline is larger than file, aborting.'
   return
endif

if numline gt 0 then nobs=numline else nobs=nlines-skipline

if skipline+nobs gt nlines then begin
   print,'RDPHALT: file is too short for skipline and numlines'
   return
endif

;Rewind file.
point_lun,unit,0

;Create the output data vectors
filename = strarr(nobs)
obj      = strarr(nobs)
fil      = strarr(nobs)
jd       = dblarr(nobs)
exptime  = fltarr(nobs)
gain     = fltarr(nobs)
rad      = fltarr(nobs)
sky1     = fltarr(nobs)
sky2     = fltarr(nobs)
serial   = intarr(nobs)
xpos     = fltarr(nobs)
ypos     = fltarr(nobs)
fwhm     = fltarr(nobs)
maxcnt   = fltarr(nobs)
mag      = fltarr(nobs)
err      = fltarr(nobs)
bad      = intarr(nobs)

jd0 = 0.0d0

; Skip leading lines (maybe)
for i=0,skipline-1 do begin
   readf,unit,line,format='(a)'
endfor

for i=0,nobs-1 do begin

; Get the next input line.
   readf,unit,line,format='(a)'

; Read the filename, object name, and filter code as string bits.
   filename[i] = gettok(line,' ')
   obj[i]      = gettok(line,"'") ; This is a dummy read to drop the first quote
   obj[i]      = gettok(line,"'")
   fil[i]      = gettok(line,' ')

; Read the rest of the data which is all numeric.
   reads,line,format='(d13.5,f9.3,f7.2,3f8.3,i5,2f9.3,f6.2,f8.1,f9.4,f8.4)', $
      jd0,exptime0,gain0,rad0,sky1_0,sky2_0,serial0, $
      xpos0,ypos0,fwhm0,maxcnt0,mag0,err0

   jd[i]       = jd0
   exptime[i]  = exptime0
   gain[i]     = gain0
   rad[i]      = rad0
   sky1[i]     = sky1_0
   sky2[i]     = sky2_0
   serial[i]   = serial0
   xpos[i]     = xpos0
   ypos[i]     = ypos0
   fwhm[i]     = fwhm0
   maxcnt[i]   = maxcnt0
   mag[i]      = mag0
   err[i]      = err0
   bad[i]      = string(strmid(line,107,2))

endfor

badio:

free_lun,unit

end
